var template = require('babel-template');

var mixinTpl = template(`
    MIXINS && mix(this, opts, MIXINS);
`);

var componentTpl = template(`
    METHODS && Object.assign(constructor.prototype, METHODS);

    ko.components.register(NAME, {
        synchronous: true,
        viewModel: {
            defaults: DEFAULTS,
            createViewModel: function (params, componentInfo) {
                return new constructor(params, componentInfo);
            }
        },
        style: STYLE,
        template: TEMPLATE,
    });
`);

module.exports = function (babel) {
    var t = babel.types;

    var ast = null;
    var constructorPath = null;

    // Captialize identifier
    var captialize = function (str) {
        if (str) {
            return str.charAt(0).toUpperCase() + str.substr(1);
        } else {
            return '';
        }
    };

    // Reset component partial ast
    var clean = function () {
        ast = Object.create(null);
        constructorPath = null;
    };

    // Render component template into ast
    var render = function () {
        return componentTpl({
            NAME: ast['name'],
            CTOR_NAME: t.Identifier(captialize(ast['name'].value)),
            DEFAULTS: ast['defaults'] || t.nullLiteral(),
            METHODS: ast['methods'] || t.nullLiteral(),
            STYLE: ast['style'] || t.stringLiteral(''),
            TEMPLATE: ast['template'] || t.stringLiteral('')
        });
    };

    // Detect whether if a valid component module
    var validateComponent = function () {
        if (!t.isStringLiteral(ast['name'])) {
            throw new Error('Can not read "name".Does this declared or exported?');
        }

        if (ast['name'] === '') {
            throw new Error('Component name is empty string.');
        }

        if (!ast['constructor'] || t.isNullLiteral(ast['constructor'])) {
            throw new Error('Can not read "constructor". Does this exported?');
        }

        return true;
    };

    var constructorVisitor = {
        FunctionExpression: {
            enter: function (path) {
                path.node.id = t.identifier(captialize(ast['name'].value));
            }
        },
        BlockStatement: {
            enter: function (path) {
                constructorPath = path;
            }
        }
    };

    var variableVisitor = {
        VariableDeclarator: {
            enter: function (path) {
                var grandPathType = path.parentPath.parentPath.type;
                var declarationName = path.node.id.name;

                if (grandPathType !== 'Program' && grandPathType !== 'ExportNamedDeclaration') {
                    return;
                }

                switch (declarationName) {
                    case 'name':
                        ast[declarationName] = path.node.init;
                        break;
                    case 'constructor':
                        ast[declarationName] = path.node.init;
                        path.traverse(constructorVisitor);
                        break;
                    default:
                        ast[declarationName] = path.node.id;
                        break;
                }
            }
        }
    };

    clean();

    return {
        visitor: {
            VariableDeclarator: variableVisitor.VariableDeclarator,

            ExportNamedDeclaration: {
                enter: function (path) {
                    path.traverse(variableVisitor);
                }
            },

            Program: {
                exit: function (path) {
                    var declaration = null;

                    if (validateComponent()) {
                        path.pushContainer('body', render());
                    }

                    if (constructorPath) {
                        constructorPath.pushContainer('body', mixinTpl({
                            MIXINS: ast['mixins'] || t.nullLiteral()
                        }));
                    }

                    clean();
                }
            }
        }
    };
};
